from typing import Optional, Tuple, Union

import cv2
import numpy as np

from cilissa.images import Image
from cilissa.operations import Transformation


class Blur(Transformation):
    """
    Blurs an image.

    Args:
        gaussian (bool): If True, uses a Gaussian filter to blur the image.
            If False, uses normalized box filter.
        kernel_size (Tuple[int, int]): Gaussian/blurring kernel size.
            Elements in tuple can differ but they both must be positive and odd.
            If using Gaussian filter they can be zero's and then they are computed from sigma.
        sigma (float): Gaussian kernel standard deviation in X direction.
            Used only with Gaussian filter.

    References:
        - https://docs.opencv.org/4.5.2/d4/d86/group__imgproc__filter.html#ga8c45db9afe636703801b0b2e440fce37
        - https://docs.opencv.org/4.5.2/d4/d86/group__imgproc__filter.html#gaabe8c836e97159a9193fb0b11ac52cf1

    """

    def __init__(self, gaussian: bool = True, kernel_size: Tuple[int, int] = (0, 0), sigma: float = 1.0) -> None:
        self.gaussian = gaussian

        self.kernel_size = kernel_size
        self.sigma = sigma

    def transform(self, image: Image) -> Image:
        im = image.as_int()

        if self.gaussian:
            new_im = cv2.GaussianBlur(im, self.kernel_size, self.sigma)
        else:
            new_im = cv2.blur(im, self.kernel_size)

        return Image(new_im)


class Sharpen(Transformation):
    """
    Sharpens an image using an unsharp mask.

    Args:
        amount (float): Amount of sharpening applied.
        threshold (int): Threshold for the low-constrast mask.
            Pixels for which the difference between the input and blurred images is less than threshold
            will remain unchanged.
        kwargs (Any): Arguments passed as kwargs will be passed to the blur transformation.

    References:
        - https://en.wikipedia.org/wiki/Unsharp_masking
    """

    def __init__(self, amount: float = 1.5, kernel_size: Tuple[int, int] = (0, 0), sigma: float = 1.0) -> None:
        # Parameters for blur
        self.kernel_size = kernel_size
        self.sigma = sigma

        # Parameters for unsharp mask
        self.amount = amount

    def transform(self, image: Image) -> Image:
        im = image.as_int()

        new_im = cv2.GaussianBlur(im, self.kernel_size, self.sigma)
        new_im = cv2.addWeighted(im, self.amount, new_im, -self.amount + 1, 0)

        return Image(new_im)


class Linear(Transformation):
    """
    Changes brightness of the image with a simple linear transformation.

    g(x) = a * f(x) + b, where `a` controls contrast and `b` controls brightness

    Brightness refers to the overall lightness or darkness of the image.
    Increasing the brightness every pixel in the frame gets lighter.

    Contrast is the difference in brightness between objects in the image.
    Increasing the contrast makes light areas lighter and dark area in the frame becomes much darker.

    Args:
        contrast (int/float/None): Value by which to change the contrast.
            1 and None is the original image.
            A float from interval (0, 1) reduces the contrast. Values above 1 increase the contrast.
        brightness (int/float/None): Value by which to change the brightness.
            0 and None is the original image.
            Negative values reduce the brightness. Positive values increase the brightness.

    References:
        - https://docs.opencv.org/3.4/d3/dc1/tutorial_basic_linear_transform.html
    """

    def __init__(
        self, contrast: Optional[Union[float, int]] = None, brightness: Optional[Union[float, int]] = None
    ) -> None:
        self.contrast = contrast
        self.brightness = brightness

    def transform(self, image: Image) -> Image:
        im = image.as_int()

        # Apply linear transformation
        new_im = cv2.convertScaleAbs(im, alpha=self.contrast, beta=self.brightness)

        return Image(new_im)


class Translation(Transformation):
    """
    Shifts an image by given amount in pixels along X and/or Y axis.

    Uses an affine transformation to perform image translation.

    Args:
        x (int): Value (in px) to move the image along X-axis.
            Positive - right, negative - left.
        y (int): Value (in px) to move the image along Y-axis.
            Positive - down, negative - up.

    References:
        - https://en.wikipedia.org/wiki/Affine_transformation
    """

    def __init__(self, x: int = 0, y: int = 0) -> None:
        self.x = x
        self.y = y

    def transform(self, image: Image) -> Image:
        im = image.as_int()

        # Transformation matrix
        M = np.array([[1, 0, self.x], [0, 1, self.y]], dtype=np.float32)
        new_im = cv2.warpAffine(im, M, (im.shape[1], im.shape[0]))

        return Image(new_im)


class Equalization(Transformation):
    """
    Constrast adjustment using histogram equalization.

    CV2 equalization is limited to 8-bit images so we use the NumPy CDF function.

    References:
        - https://en.wikipedia.org/wiki/Histogram_equalization
        - https://web.archive.org/web/20151219221513/http://www.janeriksolem.net/2009/06/histogram-equalization-with-python-and.html # noqa
    """

    def __init__(self, nbins: int = 256) -> None:
        # Number of bins for image histogram
        self.number_of_bins = nbins

    def transform(self, image: Image) -> Image:
        im = image.as_int()

        # Get normalized histogram - probability density function of each gray level
        image_histogram, bins = np.histogram(im.flatten(), self.number_of_bins, density=True)
        cdf = image_histogram.cumsum()
        cdf = 255 * cdf / cdf[-1]

        # Use linear interpolation of cdf to find new pixel values
        image_equalized = np.interp(im.flatten(), bins[:-1], cdf)
        new_im = image_equalized.reshape(im.shape).astype(np.uint8, casting="unsafe")

        return Image(new_im)



from __future__ import annotations

import base64
import logging
import os
from pathlib import Path
from typing import Optional, Tuple, Type, Union

import cv2
import numpy as np

from cilissa.classes import OrderedList
from cilissa.exceptions import NotOnImageError, ShapesNotEqual
from cilissa.roi import ROI


class Image:
    """
    `np.ndarray` wrapper, a core structure in CILISSA
    """

    path: str = ""
    name: str = ""
    im: np.ndarray

    def __init__(self, image: Union[Path, str, np.ndarray], name: Optional[str] = None) -> None:
        if name:
            self.name = name

        if isinstance(image, Path) or isinstance(image, str):
            self.load(image)
        elif isinstance(image, np.ndarray):
            self.from_array(image)
        else:
            raise TypeError("Cannot create an Image from given object")

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Image):
            comparison = self.im == o.im
        elif isinstance(o, np.ndarray):
            comparison = self.im == o
        else:
            raise TypeError(f"Cannot compare object of type Image and {type(o)}")

        return comparison.all()

    def crop(self, sl: Tuple[slice, slice]) -> Image:
        im = self.im[sl] if sl else self.im
        return Image(np.ascontiguousarray(im), name=self.name)

    @property
    def width(self) -> int:
        return self.im.shape[1]

    @property
    def height(self) -> int:
        return self.im.shape[0]

    @property
    def dtype(self) -> np.dtype:
        return self.im.dtype

    @property
    def channels_num(self) -> int:
        # 2D array is a grayscale image, 3D array gives the number of channels
        return 1 if self.im.ndim == 2 else self.im.shape[-1]

    def get_resized(self, width: Optional[int] = None, height: Optional[int] = None) -> Image:
        if width and height:
            maxsize = (width, height)
        elif width:
            maxsize = (width, int(self.get_scale_factor(width=width) * self.height))
        elif height:
            maxsize = (int(self.get_scale_factor(height=height) * self.width), height)
        else:
            return self
        return Image(cv2.resize(self.im, maxsize, interpolation=cv2.INTER_AREA), name=self.name)

    def get_scale_factor(self, width: Optional[int] = None, height: Optional[int] = None) -> float:
        if width:
            return width / self.width
        elif height:
            return height / self.height
        else:
            return 1.0

    def from_array(self, image_array: np.ndarray, at: Optional[Tuple[slice, slice]] = None) -> None:
        """
        Replaces the underlying image array with given `np.ndarray`
        """
        self.path = self.path or ""
        self.name = self.name or "Image loaded from array"

        if at:
            self.im[at] = image_array
        else:
            self.im = image_array

    def load(self, image_path: Union[Path, str]) -> None:
        """
        Loads the image from given path

        Uses cv2.imdecode instead of cv2.imread to handle unicode characters in path

        Args:
            image_path (Path/str): Path where the image is located.
        """
        self.path = str(image_path)
        self.name = os.path.basename(self.path)

        self.im = cv2.imdecode(np.fromfile(self.path, dtype=np.uint8), cv2.IMREAD_COLOR | cv2.IMREAD_ANYDEPTH)
        if self.im is None:
            raise IOError(f"Cannot open image path: `{self.path}`")

    def save(self, save_path: Union[Path, str] = "") -> None:
        """
        Saves the image

        Args:
            save_path (Path/str): Path to save the image at. Must contain the filename with extension.
                If empty string, then will save to the path the image was loaded from (if available)
        """
        if save_path:
            self.path = str(save_path)

        if self.path:
            cv2.imwrite(self.path, self.im)
        else:
            logging.error("No save path supplied!")

    def copy(self) -> Image:
        """
        Copies and returns the image
        """
        return Image(np.copy(self.im), name=self.name)

    def show(self) -> None:
        """
        Opens a CV2 window and displays the loaded image.

        Exits when user presses ESCAPE or closes window manually.
        """
        if self.im is not None:
            cv2.imshow(self.name, self.im)
            while cv2.getWindowProperty(self.name, 0) >= 0:
                k = cv2.waitKey(0)
                if k == 27:  # ESCAPE key
                    cv2.destroyWindow(self.name)
                    break

    def check_if_on_image(self, x: Optional[int] = None, y: Optional[int] = None) -> bool:
        if x and (x < 0 or x > self.width):
            return False

        if y and (y < 0 or y > self.height):
            return False

        return True

    def _as(self, data_type: Type) -> np.ndarray:
        np_type = np.result_type(self.im, data_type)
        image = np.asarray(self.im, dtype=np_type)
        return image

    def as_int(self) -> np.ndarray:
        """Converts the image to :data:`np.ndarray` of ints"""
        return self._as(np.uint8)

    def as_float(self) -> np.ndarray:
        """Converts the image to :data:`np.ndarray` of floats"""
        return self._as(np.float64)

    def as_data_uri(self) -> str:
        encoded = cv2.imencode(".png", self.im)[1]
        b64 = base64.b64encode(encoded)
        return f"data:image/png;base64,{b64.decode('ascii')}"

    def convert_to_grayscale(self) -> None:
        if self.channels_num != 3:
            logging.error(f"Cannot convert image with {self.channels_num} channels")
            return

        bgr = self.as_float()
        self.from_array(cv2.cvtColor(bgr, cv2.COLOR_BGR2GRAY))

    def __str__(self) -> str:
        return f"Image(name={self.name})"


class ImagePair:
    """
    A pair of 2 :class:`cilissa.images.Image`. Analysis is performed using this class.

    If any of the attributes in the image pair are mismatched, the attribute of the reference image
    will be used if necessary.
    """

    im1: Image
    """Reference image against which quality is measured"""
    im2: Image
    """Image whose quality is to be measured"""

    roi: Optional[ROI]
    use_roi: bool

    def __init__(self, im1: Image, im2: Image, roi: Optional[ROI] = None, use_roi: bool = True) -> None:
        self.im1 = im1
        self.im2 = im2

        if not self.matching_shape:
            raise ShapesNotEqual("Images must be of equal size to analyze")

        if not self.matching_dtype:
            logging.warn("Images have mismatched data types. Metrics will use reference image's type")

        self.use_roi = use_roi
        self.set_roi(roi)

    def __getitem__(self, key: int) -> Image:
        if key == 0:
            im = self.im1
        elif key == 1:
            im = self.im2
        else:
            raise IndexError

        slices = self._get_roi_slices()
        if slices:
            return im.crop(slices)

        return im

    def __setitem__(self, key: int, image: Image) -> None:
        if key == 0:
            self.im1.from_array(image.im, at=self._get_roi_slices())
        elif key == 1:
            self.im2.from_array(image.im, at=self._get_roi_slices())
        else:
            raise IndexError

    def __eq__(self, o: object) -> bool:
        if isinstance(o, ImagePair):
            return self.im1 == o.im1 and self.im2 == o.im2
        else:
            raise TypeError(f"Cannot compare object of type ImagePair and {type(o)}")

    def swap(self) -> None:
        """
        Swaps the reference and input images in place.
        """
        self.im1, self.im2 = self.im2, self.im1

    def copy(self) -> ImagePair:
        pair_copy = ImagePair(self.im1.copy(), self.im2.copy(), roi=self.roi, use_roi=self.use_roi)
        return pair_copy

    def set_roi(self, roi: Optional[ROI]) -> None:
        if roi and (not self.im1.check_if_on_image(roi.x0, roi.y0) or not self.im1.check_if_on_image(roi.x1, roi.y1)):
            raise NotOnImageError
        self.roi = roi

    def clear_roi(self) -> None:
        self.roi = None

    def _get_roi_slices(self) -> Optional[Tuple[slice, slice]]:
        if self.use_roi:
            return self.roi.slices if self.roi else None
        return None

    @property
    def matching_shape(self) -> bool:
        return self.im1.im.shape == self.im2.im.shape

    @property
    def matching_dtype(self) -> bool:
        return self.im1.im.dtype == self.im2.im.dtype

    def as_floats(self) -> Tuple[np.ndarray, np.ndarray]:
        """Returns a tuple with both images as :data:`np.ndarray` of floats"""
        return (self[0].as_float(), self[1].as_float())


class ImageCollection(OrderedList):
    """
    A collection of one or more :class:`cillisa.images.ImagePair`.

    Operations performed on :class:`cillisa.images.ImagePair` can be applied to the whole collection.
    """

    use_roi: bool = True

    def set_use_roi(self, value: bool) -> None:
        self.use_roi = value
        for pair in self:
            pair.use_roi = self.use_roi



from __future__ import annotations

import csv
from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import numpy as np

from cilissa.classes import Parameterized
from cilissa.images import Image
from cilissa.utils import get_parameter_display_name

if TYPE_CHECKING:
    from cilissa.operations import ImageOperation


@dataclass(frozen=True)  # type: ignore
class Result(Parameterized, ABC):
    name: str
    parameters: Dict[str, Any]
    value: Union[float, np.float64, Image]
    type: ImageOperation

    def __eq__(self, o: Any) -> bool:
        if isinstance(o, Result):
            return self.name == o.name and np.isclose(self.value, o.value, rtol=1e-05, atol=1e-08)  # type: ignore
        elif isinstance(o, float):
            return np.isclose(self.value, o, rtol=1e-05, atol=1e-08, equal_nan=False)  # type: ignore
        else:
            return self.value == o


class ResultGenerator:
    def __init__(self, results: List[Result]) -> None:
        self.results = results
        self.last_changes: Dict[str, float] = {}

    def get_change_in_metric(self, name: str, value: float) -> float:
        if name in self.last_changes:
            change = value - self.last_changes[name]
            self.last_changes[name] = change
        else:
            self.last_changes[name] = value
            return 0

        return change

    def to_html(self) -> str:
        def get_table_head() -> str:
            return """
            <thead>
                <tr>
                    <th>type</th>
                    <th>name</th>
                    <th>parameters</th>
                    <th width="128">value</th>
                    <th>change</th>
                </tr>
            </thead>
            """

        def get_table_body(results: List[Result]) -> str:
            html = ""
            for result in results:
                html += get_table_row(result)
            return html

        def get_table_row(result: Result) -> str:
            from cilissa.operations import Metric

            def get_prefix_for_change(change: Union[float, np.floating]) -> str:
                if change > 0:
                    return "+"
                else:
                    return ""

            html = "<tr>"
            html += get_table_cell(result.type.get_class_name())
            html += get_table_cell(result.name)
            html += get_table_cell(result.parameters)
            html += get_table_cell(result.value)
            if result.type == Metric:
                change = self.get_change_in_metric(result.name, float(result.value))  # type: ignore
                html += get_table_cell(change if change != 0 else "", prefix=get_prefix_for_change(change))
            html += "</tr>"
            return html

        def get_table_cell(value: Any = None, prefix: Optional[str] = None) -> str:
            def format_parameters(parameters: Dict[str, Any]) -> str:
                if parameters:
                    html = "".join(
                        [
                            "• <strong>{}</strong>: {}<br>".format(get_parameter_display_name(k), v)
                            for k, v in parameters.items()
                        ]
                    )
                else:
                    html = "No parameters"
                return html

            html = "<td align='center'>"
            if prefix:
                html += prefix

            if isinstance(value, Image):
                data_uri = value.get_resized(height=64).as_data_uri()
                html += f"<img src='{data_uri}'>"
            elif isinstance(value, dict):
                html += format_parameters(value)
            elif isinstance(value, float) or isinstance(value, np.floating):
                html += str(round(value, 4))
            else:
                html += str(value)

            html += "</td>"
            return html

        html = "<table border='1' cellpadding='8' width='728'>"
        html += get_table_head()
        html += get_table_body(self.results)
        html += "</table>"
        return html

    def to_csv(self, output_path: Union[Path, str]) -> None:
        with open(output_path, "w", newline="") as csvfile:
            writer = csv.writer(csvfile, delimiter=";", quotechar='"', quoting=csv.QUOTE_NONNUMERIC)
            writer.writerow(["type", "name", "parameters", "value"])

            for result in self.results:
                writer.writerow(
                    [
                        result.type.get_class_name(),
                        result.name,
                        str(result.parameters),
                        result.value if not isinstance(result.value, Image) else "",
                    ]
                )



import argparse

from cilissa.images import Image, ImagePair
from cilissa.operations import OperationsList
from cilissa.parsers import parse_operations_from_str, parse_roi

help_message = """
CILISSA - Interactive computer image likeness assessing.
All arguments passed after `--kwargs` flag and beginning with metric's name will be passed to metric's constructor.
Example: `cilissa -m ssim ssim-channels-num=3` will be created as `SSIM(channels_num=3)`
"""


def main() -> None:
    parser = argparse.ArgumentParser(
        description=help_message,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("-r", "--ref-image", required=True, help="Reference image against which quality is measured")
    parser.add_argument("-i", "--input-image", required=True, help="Image to be analyzed and transformed")
    parser.add_argument(
        "-O",
        "--operation",
        action="extend",
        nargs="+",
        required=True,
        help="Which operations to use (execution order depends on the order the arguments are passed)",
    )
    parser.add_argument(
        "-R", "--roi", help="ROI start and end points for analysis/transformation. Example: `0x0,384x512`"
    )
    parser.add_argument(
        "--kwargs",
        nargs="+",
        help="Keyword arguments to be passed to their respective operation. Example: `ssim-channels-num=3`",
    )
    parser.add_argument("-s", "--show-end-image", action="store_true", help="Shows the image after all transformations")
    args = parser.parse_args()

    image1 = Image(args.ref_image)
    image2 = Image(args.input_image)
    image_pair = ImagePair(image1, image2)

    if args.roi:
        roi = parse_roi(args.roi)
        image_pair.set_roi(roi)

    instances = parse_operations_from_str(args.operation or [], args.kwargs or [])
    operations = OperationsList(instances)
    result = operations.run_all(image_pair)

    if result:
        print(result)

    if args.show_end_image:
        image_pair.im2.show()


if __name__ == "__main__":
    main()



import ast
import json
import logging
from typing import Any, List, TextIO

from cilissa import metrics  # noqa
from cilissa import transformations  # noqa
from cilissa.operations import ImageOperation
from cilissa.plugins import *  # noqa
from cilissa.roi import ROI

_all_operations_dict = {op.get_class_name(): op for op in ImageOperation.get_subclasses()}


def parse_operations_from_str(operations: List[str], kwargs: List[Any]) -> List[ImageOperation]:
    """
    Parses operations and their parameters from string input.

    Parameters use the following format:

    `<operation-name>-<parameter-name>=<value>`

    where `parameter-name` uses hyphens (-) instead of underscores (_).
    """
    instances: List[ImageOperation] = []
    for op_name in operations:
        operation = _all_operations_dict.get(op_name)
        if not operation:
            continue

        op_kwargs = [arg for arg in kwargs if arg.find(operation.get_class_name()) == 0]
        parsed_kwargs = {}
        for arg in op_kwargs:
            slice_from = len(operation.get_class_name()) + 1
            value: Any
            try:
                # Checking if argument has a value supplied
                key = arg[slice_from : arg.index("=")].replace("-", "_")
                value = arg[arg.index("=") + 1 :]
            except ValueError:
                # Argument is a flag
                key = arg[slice_from:]
                value = True

            if isinstance(value, str):
                # Try to figure out the correct type for argument
                try:
                    value = ast.literal_eval(value)
                except ValueError:
                    # Argument is a string or cannot guess correct type
                    pass

            parsed_kwargs[key] = value

        instance = operation(**parsed_kwargs)  # type: ignore
        instances.append(instance)

    return instances


def parse_operations_from_json(fp: TextIO) -> List[ImageOperation]:
    """
    Parses operations and their parameters from a JSON file.

    Expected dictionary structure:
    `[{"name": "ssim", "parameters": {"channels_num": 3, "sigma": 1.5}}]`
    """
    data = json.load(fp)

    instances: List[ImageOperation] = []
    try:
        for operation in data:
            cls = _all_operations_dict.get(operation["name"])
            if not cls:
                continue

            instance = cls()
            for param, value in operation["parameters"].items():
                instance.set_parameter(param, value)

            instances.append(instance)
    except (KeyError, TypeError):
        logging.error("Malformed JSON file supplied")
        return []

    return instances


def parse_roi(string: str) -> ROI:
    """
    Parses ROI from string.

    Expected string format:
    `0x0,512x512` (width x height)
    """
    points = string.split(",")

    start_point = points[0].split("x")
    end_point = points[1].split("x")

    try:
        x0 = int(start_point[0])
        y0 = int(start_point[1])
        x1 = int(end_point[0])
        y1 = int(end_point[1])
    except TypeError:
        logging.error("ROI points must be integers")

    return ROI(x0, y0, x1, y1)



import inspect
from abc import ABC
from typing import Any, Dict, Iterator, List, Optional, Tuple, Type, Union


class Parameterized(ABC):
    def get_parameters_dict(self) -> Dict[str, Any]:
        d = vars(self)
        return d

    def get_parameter_type(self, parameter: str) -> Tuple[Union[Type, None], bool]:
        """
        Gets class variable's type from the `__init__` function.

        Works best if the function is type hinted, fallbacks to variable's value type.
        """
        signature = inspect.signature(self.__init__)  # type: ignore
        parameter_obj = signature.parameters.get(parameter, None)
        if parameter_obj:
            annotation = parameter_obj.annotation
            try:
                if annotation.__origin__ == Union:
                    t = annotation.__args__
                    optional = type(None) in t
                    return t[0], optional
            except AttributeError:
                t = annotation
                return t, False

        attr = self.get_parameter(parameter, None)
        if attr is not None:
            return type(attr), False

        return None, True

    def set_parameter(self, parameter: str, value: Any) -> None:
        setattr(self, parameter, value)

    def get_parameter(self, parameter: str, default: Optional[Any] = None) -> Any:
        return getattr(self, parameter, default)


class OrderedList:
    """
    Simple list that can be reordered.

    Behaviour similar to `list`, but does not derive from it to avoid lay-out conflict with QObject

    Implements `get_order` and `change_order` methods to manage the list's order.
    """

    def __init__(self, items: List[Any] = []) -> None:
        self.items = list(items)

    def __str__(self) -> str:
        return f"OrderedList(items=[{', '.join([item.get_display_name() for item in self.items])}])"

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.items)

    def __getitem__(self, index: int) -> Any:
        return self.items[index]

    def __setitem__(self, index: int, value: Any) -> None:
        self.items[index] = value

    @property
    def is_empty(self) -> bool:
        return len(self) == 0

    def push(self, item: Any) -> None:
        self.items.append(item)

    def pop(self, index: int = -1) -> Any:
        return self.items.pop(index)

    def get_order(self) -> List[Tuple[int, Any]]:
        return [(index, element) for index, element in enumerate(self.items)]

    def change_order(self, a: int, b: int) -> None:
        self.items[a], self.items[b] = self.items[b], self.items[a]

    def clear(self) -> None:
        self.items = []



from __future__ import annotations

from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from copy import deepcopy
from typing import Any, List, Type, Union

from cilissa.classes import OrderedList, Parameterized
from cilissa.images import Image, ImageCollection, ImagePair
from cilissa.results import Result


class ImageOperation(Parameterized, ABC):
    """
    Base class for all operations that can be performed on an image.

    Display name and name used in various dicts is deduced from the class name.
    """

    @classmethod
    def get_subclasses(cls) -> List[Type[ImageOperation]]:
        subclasses = []

        for subclass in cls.__subclasses__():
            subclasses.append(subclass)
            subclasses.extend(subclass.get_subclasses())

        return subclasses

    @classmethod
    def get_class_name(cls) -> str:
        return cls.__name__.lower()

    @classmethod
    def get_display_name(cls) -> str:
        return cls.get_class_name().upper()

    @abstractmethod
    def run(self, image_pair: ImagePair) -> Result:
        raise NotImplementedError("All ImageOperation subclasses must implement the `run` method")

    def generate_result(self, **kwargs: Any) -> Result:
        return Result(name=self.get_display_name(), parameters=deepcopy(self.get_parameters_dict()), **kwargs)


class OperationsList(OrderedList):
    def run_all(self, images: Union[ImagePair, ImageCollection]) -> Any:
        if isinstance(images, ImagePair):
            pair_copy = images.copy()
            res = self._use_operations_on_pair(pair_copy)
            return res

        elif isinstance(images, ImageCollection):
            with ThreadPoolExecutor(max_workers=None) as executor:
                futures = {
                    executor.submit(self._use_operations_on_pair, pair.copy()): index
                    for index, pair in images.get_order()
                }

                results = {}
                for future in as_completed(futures):
                    results[futures[future]] = future.result()

            # Sort results to their original order
            return [result[1] for result in sorted(results.items())]
        else:
            raise TypeError("Objects must be of type: ImagePair, ImageCollection")

    def _use_operations_on_pair(self, image_pair: ImagePair) -> List[Result]:
        results = []
        for operation in self:
            result = operation.run(image_pair)
            if result is not None:
                results.append(result)

        return results


class Transformation(ImageOperation, ABC):
    """
    Base class for creating new transformations to use in the program.
    """

    @classmethod
    def get_display_name(cls) -> str:
        return cls.get_class_name().capitalize()

    @abstractmethod
    def transform(self, image: Image) -> Image:
        pass

    def run(self, image_pair: ImagePair) -> Result:
        image = image_pair[1]
        transformed_image = self.transform(image)
        image_pair[1] = transformed_image
        return self.generate_result(type=Transformation, value=transformed_image)


class Metric(ImageOperation, ABC):
    """
    Base class for creating new metrics to use in the program.
    """

    @abstractmethod
    def analyze(self, image_pair: ImagePair) -> float:
        pass

    def run(self, image_pair: ImagePair) -> Result:
        value = self.analyze(image_pair)
        return self.generate_result(type=Metric, value=value)



import glob
from os.path import basename, dirname, isfile, join

modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")]



from typing import Optional

import cv2
import numpy as np

from cilissa.helpers import apply_to_channels, crop_array, get_max_pixel_value
from cilissa.images import ImagePair
from cilissa.operations import Metric


class MSE(Metric):
    """
    Mean squared error (MSE)

    Average squared difference between the estimated values and the actual value.

    References:
        - https://en.wikipedia.org/wiki/Mean_squared_error
    """

    def __init__(self, rmse: bool = False) -> None:
        self.root_mean_square_error = rmse

    def analyze(self, image_pair: ImagePair) -> float:
        im1, im2 = image_pair.as_floats()
        result = np.mean(np.square((im1 - im2)), dtype=np.float64)
        if self.root_mean_square_error:
            result = np.sqrt(result)
        return result


class PSNR(Metric):
    """
    Peak signal-to-noise ratio (PSNR)

    Ratio between the maximum possible power of a signal and
    the power of corrupting noise that affects the fidelity of its representation.

    References:
        - https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio
    """

    def __init__(self, max_pixel_value: Optional[int] = None) -> None:
        # Maximum possible pixel value of the image
        self.max_pixel_value = max_pixel_value

    def analyze(self, image_pair: ImagePair) -> float:
        dmax = self.max_pixel_value or get_max_pixel_value(image_pair[0].dtype)

        err = MSE().analyze(image_pair)
        if err == 0:
            result = np.inf
        else:
            result = 20 * np.log10(dmax) - 10 * np.log10(err)
        return result


class SSIM(Metric):
    """
    Structural similarity index measure (SSIM)

    The SSIM Index quality assessment index is based on the computation of three terms,
    namely the luminance term, the contrast term and the structural term.
    The overall index is a multiplicative combination of the three terms.

    Args:
        channels_num (int/None): If None, image is assumed to be grayscale (single channel).
            Otherwise the number of channels should be specified here.

    Returns:
        float: Overall quality measure of the entire image (MSSIM)

    References:
        - https://en.wikipedia.org/wiki/Structural_similarity
        - https://ece.uwaterloo.ca/~z70wang/publications/ssim.pdf
    """

    def __init__(
        self,
        channels_num: Optional[int] = None,
        sigma: float = 1.5,
        truncate: float = 3.5,
        K1: float = 0.01,
        K2: float = 0.03,
    ) -> None:
        # Number of channels in image
        self.channels_num = channels_num

        # Small constants 0 <= K1, K2 <= 1
        self.K1 = K1
        self.K2 = K2
        if self.K1 < 0:
            raise ValueError("K1 must be positive!")
        if self.K2 < 0:
            raise ValueError("K2 must be positive!")

        # Standard deviation for weighting function, 0 < sigma
        self.sigma = sigma
        if self.sigma < 0:
            raise ValueError("Sigma must be positive!")

        # Truncate the Gaussian filter at this many standard deviations, 0 < truncate
        self.truncate = truncate
        if self.truncate < 0:
            raise ValueError("Truncate must be positive!")

    def mssim_single_channel(self, im1: np.ndarray, im2: np.ndarray) -> float:
        dmax = im1.max()
        dmin = im1.min()
        drange = dmax - dmin

        # Compute weighted means using Gaussian weighting function
        ux = cv2.GaussianBlur(im1, (0, 0), self.sigma)
        uy = cv2.GaussianBlur(im2, (0, 0), self.sigma)

        # Compute weighted variances and covariances
        uxx = cv2.GaussianBlur(im1 ** 2, (0, 0), self.sigma)
        uyy = cv2.GaussianBlur(im2 ** 2, (0, 0), self.sigma)
        uxy = cv2.GaussianBlur(im1 * im2, (0, 0), self.sigma)
        vx = uxx - ux * ux
        vy = uyy - uy * uy
        vxy = uxy - ux * uy

        # Constants to avoid instability when ux**2 + uy**2 are close to zero (formula 7)
        L = drange
        C1 = (self.K1 * L) ** 2
        C2 = (self.K2 * L) ** 2

        # Final form of the SSIM index (formula 13, page 605)
        A1, A2, B1, B2 = (2 * ux * uy + C1, 2 * vxy + C2, ux ** 2 + uy ** 2 + C1, vx + vy + C2)
        D = B1 * B2
        S = (A1 * A2) / D

        # Avoid edge effects by ignoring filter radius around edges
        # Pad equal to radius as in scipy gaussian_filter
        # https://github.com/scipy/scipy/blob/v1.7.0/scipy/ndimage/filters.py#L258
        pad = int(self.truncate * self.sigma + 0.5)

        return crop_array(S, pad).mean()

    def analyze(self, image_pair: ImagePair) -> float:
        im1, im2 = image_pair.as_floats()

        ch_num = self.channels_num or image_pair[0].channels_num

        result = apply_to_channels(im1, im2, self.mssim_single_channel, ch_num)
        return result


class UIQI(Metric):
    """
    Universal Image Quality Index (UIQI)

    Combines loss of correlation, luminance distortion and contrast distortion.
    Predecessor of SSIM metric.

    References:
        - https://ece.uwaterloo.ca/~z70wang/publications/quality_2c.pdf
    """

    def __init__(self, channels_num: Optional[int] = None, block_size: int = 8) -> None:
        self.channels_num = channels_num

        # Sliding window size
        self.block_size = block_size

    def uiqi_single_channel(self, im1: np.ndarray, im2: np.ndarray) -> float:
        N = self.block_size ** 2

        im1_sq = im1 ** 2
        im2_sq = im2 ** 2
        im12 = im1 * im2

        unif_filter = np.ones(self.block_size, np.float32) / (self.block_size ** 2)

        im1_sum = cv2.filter2D(im1, ddepth=-1, kernel=unif_filter)
        im2_sum = cv2.filter2D(im2, ddepth=-1, kernel=unif_filter)
        im1_sq_sum = cv2.filter2D(im1_sq, ddepth=-1, kernel=unif_filter)
        im2_sq_sum = cv2.filter2D(im2_sq, ddepth=-1, kernel=unif_filter)
        im12_sum = cv2.filter2D(im12, ddepth=-1, kernel=unif_filter)

        im12_sum_mul = im1_sum * im2_sum
        im12_sq_sum_mul = im1_sum ** 2 + im2_sum ** 2
        numerator = 4 * (N * im12_sum - im12_sum_mul) * im12_sum_mul
        denominator1 = N * (im1_sq_sum + im2_sq_sum) - im12_sq_sum_mul
        denominator = denominator1 * im12_sq_sum_mul

        q_map = np.ones(denominator.shape)
        index = np.logical_and((denominator1 == 0), (im12_sq_sum_mul != 0))
        q_map[index] = (2 * im12_sum_mul[index]) / (im12_sq_sum_mul[index])
        index = denominator != 0
        q_map[index] = numerator[index] / denominator[index]

        s = int(self.block_size / 2)
        return np.mean(q_map[s:-s, s:-s])

    def analyze(self, image_pair: ImagePair) -> float:
        im1, im2 = image_pair.as_floats()

        ch_num = self.channels_num or image_pair[0].channels_num

        result = apply_to_channels(im1, im2, self.uiqi_single_channel, ch_num)
        return result


class VIFP(Metric):
    """
    Pixel Based Visual Information Fidelity (VIF-P)

    Employs natural scene statistical (NSS) models in conjunction with a distortion (channel) model to quantify
    the information shared between the test and the reference images.

    The pixel domain algorithm is not described in the paper below. This is a computationally simpler
    derivative of the algorithm presented in the paper, based on the original Matlab implementation.

    References:
        - http://live.ece.utexas.edu/publications/2004/hrs_ieeetip_2004_imginfo.pdf
        - https://github.com/utlive/vif_pixel
    """

    def __init__(self, channels_num: Optional[int] = None, sigma: float = 2.0) -> None:
        self.channels_num = channels_num

        # Variance of the visual noise
        self.sigma = sigma

    def vifp_single_channel(self, im1: np.ndarray, im2: np.ndarray) -> float:
        eps = 1e-10
        numerator = 0
        denominator = 0

        for scale in range(1, 5):
            N = 2 ** (4 - scale + 1) + 1
            sd = N / 5.0

            if scale > 1:
                im1 = cv2.GaussianBlur(im1, (0, 0), sd)
                im2 = cv2.GaussianBlur(im2, (0, 0), sd)

            mu1 = cv2.GaussianBlur(im1, (0, 0), sd)
            mu2 = cv2.GaussianBlur(im2, (0, 0), sd)
            mu1_sq = mu1 ** 2
            mu2_sq = mu2 ** 2
            mu12 = mu1 * mu2

            sigma1_sq = cv2.GaussianBlur(im1 ** 2, (0, 0), sd) - mu1_sq
            sigma2_sq = cv2.GaussianBlur(im2 ** 2, (0, 0), sd) - mu2_sq
            sigma12 = cv2.GaussianBlur(im1 * im2, (0, 0), sd) - mu12

            sigma1_sq[sigma1_sq < 0] = 0
            sigma2_sq[sigma2_sq < 0] = 0

            g = sigma12 / (sigma1_sq + eps)
            sv_sq = sigma2_sq - (g * sigma12)

            g[sigma1_sq < eps] = 0
            sv_sq[sigma1_sq < eps] = sigma2_sq[sigma1_sq < eps]
            sigma1_sq[sigma1_sq < eps] = 0

            g[sigma2_sq < eps] = 0
            sv_sq[sigma2_sq < eps] = 0

            sv_sq[g < 0] = sigma2_sq[g < 0]
            g[g < 0] = 0
            sv_sq[sv_sq <= eps] = eps

            numerator += np.sum(np.log10(1 + g ** 2 * (sigma1_sq / (sv_sq + self.sigma))))
            denominator += np.sum(np.log10(1 + (sigma1_sq / self.sigma)))

        return numerator / denominator

    def analyze(self, image_pair: ImagePair) -> float:
        im1, im2 = image_pair.as_floats()

        ch_num = self.channels_num or image_pair[0].channels_num

        result = apply_to_channels(im1, im2, self.vifp_single_channel, ch_num)
        return result


class SAM(Metric):
    """
    Spectral Angle Mapper (SAM)

    Physically-based spectral classification that uses an n-D angle to match pixels to reference spectra.

    This technique is relatively insensitive to illumination and albedo effects.

    References:
        - https://ntrs.nasa.gov/citations/19940012238
    """

    def __init__(self, to_degrees: bool = True) -> None:
        self.convert_to_degrees = to_degrees

    def analyze(self, image_pair: ImagePair) -> float:
        im1, im2 = image_pair.as_floats()

        # Spectral angles are first computed for each pair of pixels
        numerator = np.sum(im1 * im2, axis=2)
        denominator = np.linalg.norm(im1, axis=2) * np.linalg.norm(im2, axis=2)
        val = np.clip(numerator / denominator, -1, 1)
        sam_angles = np.arccos(val)

        if self.convert_to_degrees:
            sam_angles *= 180 / np.pi

        return np.mean(np.nan_to_num(sam_angles))



from dataclasses import dataclass
from typing import Tuple


@dataclass
class ROI:
    x0: int
    y0: int
    x1: int
    y1: int

    @property
    def start_point(self) -> Tuple[int, int]:
        return (self.x0, self.y0)

    @property
    def end_point(self) -> Tuple[int, int]:
        return (self.x1, self.y1)

    @property
    def slices(self) -> Tuple[slice, slice]:
        return (slice(self.y0, self.y1), slice(self.x0, self.x1))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cilissa', 'cilissa.plugins']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.20.3,<2.0.0', 'opencv-python>=4.5.2,<5.0.0']

extras_require = \
{'gui': ['PySide6>=6.1.2,<7.0.0']}

entry_points = \
{'console_scripts': ['cilissa = cilissa.__main__:main']}

setup_kwargs = {
    'name': 'cilissa',
    'version': '0.7.1',
    'description': 'Interactive tool for assessing digital image similarity',
    'long_description': '<p align="center">\n    <img src="https://raw.githubusercontent.com/exler/CILISSA/main/docs/_static/logo.png" width="328">\n</p>\n<p align="center">\n    <strong>C</strong>omputer <strong>I</strong>mage <strong>Li</strong>keness A<strong>ss</strong>essing <strong>A</strong>utomation\n</p>\n<p align="center">\n    <!-- Badges -->\n    <img src="https://github.com/exler/CILISSA/actions/workflows/quality.yml/badge.svg">\n    <img src="https://github.com/exler/CILISSA/actions/workflows/tests.yml/badge.svg">\n    <a href="https://codecov.io/gh/exler/CILISSA">\n        <img src="https://codecov.io/gh/exler/CILISSA/branch/main/graph/badge.svg?token=Dixb5buMQr"/>\n    </a>\n    <a href="https://cilissa.readthedocs.io/en/latest/">\n        <img src="https://img.shields.io/readthedocs/cilissa">\n    </a>    \n</p>\n\n## Overview\n\nCILISSA allows for the use of various metrics to perform full-reference image comparisons.\n\nIt features the most popular full-reference image quality metrics, image transformations and translations. \nCILISSA is also very extensible and new operations can be easily added.\n\nCILISSA has an optional Qt-based graphical interface that lets you experiment with various operations, their orders and properties.\n\n## Requirements\n\n* Python >= 3.7\n\n## Installation\n### Install \n```bash\n$ pip install cilissa\n```\n\n### Releases\n\nBinaries for Windows and Linux can be found in [GitHub releases](https://github.com/exler/CILISSA/releases).\n\n## Usage\n\n### GUI\n\nInformation about the GUI can be found in the [cilissa_gui/README.md](cilissa_gui/README.md) file.\n\n### CLI\n\nCurrently the CLI only supports working with a single pair of images.\n\nThe parameters of metrics and transformations can be modified by passing them to the `--kwargs` argument using the following format:\n```\n<operation-name>-<parameter-name>=<value>\n``` \nwhere `parameter-name` uses hyphens (-) instead of underscores (_)\n\n## Documentation\n\nDocumentation is hosted on [Read the Docs](https://cilissa.readthedocs.io/).\n',
    'author': 'Kamil Marut',
    'author_email': 'kamil@kamilmarut.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/exler/CILISSA',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



from typing import Callable

import numpy as np


def get_max_pixel_value(dtype: np.dtype) -> int:
    if dtype == np.uint8:
        return 255
    elif dtype == np.uint16:
        return 65535
    elif dtype == np.uint32:
        return 2147483647
    elif np.issubdtype(dtype, np.floating):
        return 1
    else:
        raise ValueError(f"Cannot get max pixel value for dtype {dtype}")


def apply_to_channels(im1: np.ndarray, im2: np.ndarray, func: Callable, channels_num: int) -> float:
    results = np.empty(channels_num)
    for channel in range(channels_num):
        channel_result = func(im1[:, :, channel], im2[:, :, channel])
        results[channel] = channel_result

    return np.mean(np.nan_to_num(results))


def crop_array(array: np.ndarray, crop_width: int) -> np.ndarray:
    """
    Crop input `array` by `crop_width` in every dimension
    """
    array = np.array(array, copy=False)

    crops = [[crop_width, crop_width]] * array.ndim
    slices = tuple(slice(a, array.shape[i] - b) for i, (a, b) in enumerate(crops))
    cropped = array[slices]

    return cropped



class ShapesNotEqual(Exception):
    """Two images must be of equal size to compare"""

    pass


class NotOnImageError(Exception):
    """Points of interest are outside image's boundary"""

    pass
